@@ -31,6 +31,7 @@
 #include "http_log.h"
 
 #include <iconv.h>
+#include <pwd.h>
 
 #ifndef _UNISTD_H
 #include <unistd.h>
@@ -52,7 +53,7 @@
  * URL:
  *   http://modurl.kldp.net/
  *
- * $Id: mod_url.c,v 1.8 2007/06/11 08:02:14 oops Exp $
+ * $Id: mod_url.c,v 1.11 2007/11/09 18:25:58 oops Exp $
  */
 
 /*
@@ -468,11 +469,14 @@ static int check_redurl (request_rec * r)
 			"  Fixed => %s", c_uri);
 
 	/*
-	 * Full path and uri have character that is out of range ascii.
-	 * But, encoding is different each other. So, mod_url sending
-	 * 301 HTTP_MOVED_PERMANENTLY with converted URL
+	 * Content type is null and full path and uri have character that is
+	 * out of range ascii. But, encoding is different each other. So,
+	 * mod_url sending 301 HTTP_MOVED_PERMANENTLY with converted URL
 	 */
-	if ( ric_r == REDURL_ICONV_FALSE ) {
+	if ( ric_r == REDURL_ICONV_FALSE || r->content_type == NULL ) {
+		check_redurl_iconv_free (ric);
+		free (realpath);
+
 		/* working ready */
 		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
 				"  Goto => %s", ap_construct_url (r->pool, c_uri, r));
@@ -513,6 +517,8 @@ static int check_redurl (request_rec * r)
 	apr_stat (&r->finfo, r->filename, APR_FINFO_MIN, r->pool);
 	check_redurl_iconv_free (ric);
 
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
+			"  finfo.filetype     => %d", r->finfo.filetype);
 	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
 			"  r->uri             => %s", r->uri);
 	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
@@ -523,12 +529,280 @@ static int check_redurl (request_rec * r)
 			"  r->filename        => %s", r->filename);
 	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
 			"  r->canonical_filename => %s", r->canonical_filename);
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
+			"  r->content-type    => %s", r->content_type);
 
 	return OK;
 }
 
+static int pre_redurl (request_rec * r) {
+	urlconfig	* cfg;
+	iconv_s		* ric;
+	char		* realpath, * enc;
+	char		  docroot[256] = { 0, };
+	short		  urilen, rootlen, real_len, ric_r;
+	struct		  stat realstat;
+	short		  is_userdir = 0;
+
+	cfg = ap_get_module_config (r->per_dir_config, &redurl_module);
+	if ( ! cfg->enabled )
+		return DECLINED;
+
+	/* We've already got a file of some kind or another */
+	if ( r->proxyreq || (r->finfo.filetype != 0) )
+		return DECLINED;
+
+	/* This is a sub request - don't mess with it */
+	if ( r->main )
+		return DECLINED;
+
+	/*
+	 * Removes double or multiple slashes from a r->uri
+	 */
+	ap_no2slash (r->uri);
+
+	/* Account access : /~user */
+	if ( r->uri[0] == '/' && r->uri[1] == '~' ) {
+#ifndef URL_NOUSERDIR
+#if defined(WIN32) || defined(NETWARE)
+		/* Need to figure out home dirs on NT and NetWare 
+		 * Shouldn't happen here, though, we trap for this in set_user_dir
+		 */
+		return DECLINED;
+#else
+		extern module AP_MODULE_DECLARE_DATA userdir_module;
+		typedef struct userdir_config {
+			int globally_disabled;
+			char *userdir;
+			apr_table_t *enabled_users;
+			apr_table_t *disabled_users;
+		} userdir_config;
+
+		const userdir_config *ucfg;
+		const char *userdirs;
+		struct passwd *pw;
+		const char *w, *dname;
+
+		is_userdir = 1;
+
+		ucfg = (userdir_config *)
+				ap_get_module_config (r->server->module_config, &userdir_module);
+		userdirs = ucfg->userdir;
+
+		if ( ucfg->userdir == NULL )
+			return DECLINED;
+
+		dname = r->uri + 2;
+		w = ap_getword (r->pool, &dname, '/');
+
+		/*
+		 * The 'dname' funny business involves backing it up to capture the '/'
+		 * delimiting the "/~user" part from the rest of the URL, in case there
+		 * was one (the case where there wasn't being just "GET /~user HTTP/1.0",
+		 * for which we don't want to tack on a '/' onto the filename).
+		 */
+		if (dname[-1] == '/') {
+			--dname;
+		}
+		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, "UDir w => %s", w);
+		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, "UDir dname => %s", dname);
+
+		/*
+		 * If there's no username, it's not for us.  Ignore . and .. as well.
+		 */
+		if ((w[0] == '\0') || ((w[1] == '.') &&
+			((w[2] == '\0') || ((w[2] == '.') && (w[3] == '\0'))))) {
+			return DECLINED;
+		}
+
+		/*
+		 * Nor if there's an username but it's in the disabled list.
+		 */
+		if (apr_table_get(ucfg->disabled_users, w) != NULL)
+			return DECLINED;
+
+		/*
+		 * If there's a global interdiction on UserDirs, check to see if this
+		 * name is one of the Blessed.
+		 */
+		if (ucfg->globally_disabled
+				&& (apr_table_get(ucfg->enabled_users, w) == NULL))
+			return DECLINED;
+
+		while ( *userdirs) {
+			const char * userdir;
+			int is_absolute;
+
+			userdir = ap_getword_conf (r->pool, &userdirs);
+			is_absolute = ap_os_is_path_absolute (r->pool, userdir);
+
+			if ( strchr (userdir, '*') ) {
+				char *x = ap_getword (r->pool, &userdir, '*');
+
+				if ( is_absolute ) {
+					/* token '*' within absolute path
+					 * serves [UserDir arg-pre*][user][UserDir arg-post*]
+					 * /somepath/ * /somedir + /~smith -> /somepath/smith/somedir
+					 */
+					sprintf (docroot, "%s%s%s%s", x, w, userdir, dname);
+				} else {
+					return DECLINED;
+				}
+			} else if ( is_absolute ) {
+				/* An absolute path, no * token:
+				 * serves [UserDir arg]/[user]
+				 * /home + /~smith -> /home/smith
+				 */
+				if ( userdir[strlen (userdir) - 1] == '/' )
+					sprintf (docroot, "%s%s%s", userdir, w, dname);
+				else
+					sprintf (docroot, "%s/%s%s", userdir, w, dname);
+			} else if ( strchr (userdir, ':') ) {
+				return DECLINED;
+			} else {
+				struct passwd *pw;
+
+				if ( (pw = getpwnam (w)) ) {
+#ifdef OS2
+					/* Need to manually add user anem for OS/2 */
+					sprintf (docroot, "%s%s/%s%s", pw->pw_dir, w, userdir, dname);
+#else
+					sprintf (docroot, "%s/%s%s", pw->pw_dir, userdir, dname);
+#endif
+				}
+			}
+		}
+#endif
+#else
+		return DECLINED;
+#endif /* URL_NOUSERDIR */
+	} else {
+		strcpy (docroot, ap_document_root (r));
+	}
+	rootlen = strlen (docroot);
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, "UDir docroot => %s", docroot);
+
+	/* removed last slash of path */
+	if ( rootlen && docroot[rootlen - 1] == '/' ) {
+		docroot[rootlen - 1] = 0;
+		rootlen -= 1;
+	}
+
+	urilen  = strlen (r->uri);
+	real_len = rootlen + urilen;
+
+	if ( (realpath = (char *) malloc (sizeof (char) * (real_len + 1))) == NULL ) {
+		redurl_mem_error (r, APLOG_MARK, "pre realpath");
+		return DECLINED;
+	}
+
+	if ( is_userdir ) {
+		strcpy (realpath, docroot);
+	} else {
+		strcpy (realpath, rootlen ? (const char *) docroot : "");
+		strcat (realpath, urilen  ? (const char *) r->uri : "");
+	}
+
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, "pre r->uri   => %s", r->uri);
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, "pre realpath => %s", realpath);
+
+	/* exists and skip pre_redurl */
+	if ( stat (realpath, &realstat) > -1 ) {
+		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
+				"    .. Real file is exist. Skip now.");
+		free (realpath);
+		return DECLINED;
+	}
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
+			"    .. Real file is not exist.");
+
+	/* convert real path */
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
+			"Pre Real path Converting"); 
+
+	if ( (ric = (iconv_s *) malloc (sizeof (iconv_s) + 1)) == NULL ) {
+		redurl_mem_error (r, APLOG_MARK, "pre real: ric");
+		free (realpath);
+
+		return DECLINED;
+	}
+	check_redurl_iconv (r, cfg, ric, realpath);
+	ric_r = check_redurl_iconv_result (ric);
+
+	if ( ric_r == REDURL_ICONV_FALSE || ! strcmp (realpath, ric->uri) ) {
+		check_redurl_iconv_free (ric);
+		free (realpath);
+		return ric_r ? DECLINED : OK;
+	}
+
+	/* skip cause of no exist convert file */
+	if ( stat (ric->uri, &realstat) < 0 ) {
+		ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
+				"  Not Found => pre %s", ric->uri);
+
+		check_redurl_iconv_free (ric);
+		return DECLINED;
+	}
+
+	free (realpath);
+	realpath = strdup (ric->uri);
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
+			"pre realpath           => %s", realpath);
+
+	check_redurl_iconv_free (ric);
+
+	/* convert init uri */
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r, "Pre URI Converting");
+
+	if ( (ric = (iconv_s *) malloc (sizeof (iconv_s) + 1)) == NULL ) {
+		redurl_mem_error (r, APLOG_MARK, "pre uri: ric");
+		free (realpath);
+		return DECLINED;
+	}
+
+	check_redurl_iconv (r, cfg, ric, r->uri);
+	ric_r = check_redurl_iconv_result (ric);
+
+	if ( ric_r == REDURL_ICONV_FALSE || ! strcmp (r->uri, ric->uri) ) {
+		check_redurl_iconv_free (ric);
+		free (realpath);
+		return ric_r ? DECLINED : OK;
+	}
+
+	if ( (enc = check_redurl_encode (ric->uri, strlen (ric->uri), NULL)) == NULL ) {
+		redurl_mem_error (r, APLOG_MARK, "enc");
+		check_redurl_iconv_free (ric);
+		free (realpath);
+
+		return DECLINED;
+	}
+
+	r->uri = apr_pstrdup (r->pool, ric->uri);
+	r->unparsed_uri = apr_pstrdup (r->pool, enc);
+	r->parsed_uri.path = apr_pstrdup (r->pool, r->uri);
+	free (enc);
+
+	apr_stat (&r->finfo, realpath, APR_FINFO_MIN, r->pool);
+
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
+			"pre finfo.filetype     => %d", r->finfo.filetype);
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
+			"pre realpath           => %s", realpath);
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
+			"pre r->uri             => %s", r->uri);
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
+			"pre r->unparsed_uri    => %s", r->unparsed_uri);
+	ap_log_rerror(APLOG_MARK, APLOG_DEBUG, APR_SUCCESS, r,
+			"pre r->parsed_uri.path => %s", r->parsed_uri.path);
+
+	free (realpath);
+
+	return DECLINED;
+}
+
 static void register_hooks (apr_pool_t * p)
 {
+	ap_hook_translate_name (pre_redurl, NULL, NULL, APR_HOOK_LAST);
 	ap_hook_fixups (check_redurl, NULL, NULL, APR_HOOK_LAST);
 }
 
